
  if (ntria > 0) then

  boundary_element_loop : do i = 1, ntria

    ib = tria(i,4)

    nodes_of_triangle : do j = 1, 3 ! Nodes: nj, nr, nl

    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------
    ! Second-order accurate boundary flux quadratue for tetra (6/8,1/8,1/8)
    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------
     !----------------------------------------------------------------
     ! Node j: Phi_jb(bn): BC is implemented here.

          nj = tria( i, j )

       call set_bc_state( u(nj,1),u(nj,2),u(nj,3),u(nj,4),ib,bn(i,:),x(nj),y(nj),z(nj), ub,qnb) !Get ub,qnb
       call get_right_state(  bc_type(ib),u(nj,:),ub,qnb, bn(i,:),  duRduL,uR )                 !Get duR/duL, uR
       call interface_jac_b( u(nj,:), uR, bn(i,:), sqrt(bn_mag(i))/ar(nj), dFnduL, dFnduR )

     !Contribution to node j (diagonal block):
      jac_diag(nj,:,:) = jac_diag(nj,:,:) &
                       + 6.0_p2/8.0_p2*(dFnduL + matmul(dFnduR,duRduL)) * (third*bn_mag(i))

     !----------------------------------------------------------------
     ! Flux between j and r: Phi_jr(bn): BC is implemented here.

          nr = tria( i, loc_nghbr(j,1) )

       call set_bc_state( u(nr,1),u(nr,2),u(nr,3),u(nr,4),ib,bn(i,:),x(nr),y(nr),z(nr), ub,qnb) !Get ub,qnb
       call get_right_state(  bc_type(ib),u(nr,:),ub,qnb, bn(i,:),  duRduL,uR )                 !Get duR/duL, uR
       call interface_jac_b( u(nr,:), uR, bn(i,:), sqrt(bn_mag(i))/ar(nr), dFnduL, dFnduR )

     !Contribution to node j (off-diagonal block):
      k = kth_nghbr(nj,nr)
      jac_off( k0(nj)+k ,:,:) = jac_off( k0(nj)+k, :,:) &
                              + 1.0_p2/8.0_p2*(dFnduL + matmul(dFnduR,duRduL)) * (third*bn_mag(i))

     !----------------------------------------------------------------
     ! Flux between j and l: Phi_jl(bn): BC is implemented here.

          nl = tria( i, loc_nghbr(j,2) )

       call set_bc_state( u(nl,1),u(nl,2),u(nl,3),u(nl,4),ib,bn(i,:),x(nl),y(nl),z(nl), ub,qnb) !Get ub,qnb
       call get_right_state(  bc_type(ib),u(nl,:),ub,qnb, bn(i,:),  duRduL,uR )                 !Get duR/duL, uR
       call interface_jac_b( u(nl,:), uR, bn(i,:), sqrt(bn_mag(i))/ar(nl), dFnduL, dFnduR )

     !Add the boundary contribution to the residual with the 2nd-order formula:
     !Contribution to node j (off-diagonal block):
      k = kth_nghbr(nj,nl)
      jac_off( k0(nj)+k ,:,:) = jac_off( k0(nj)+k, :,:) &
                              + 1.0_p2/8.0_p2*(dFnduL + matmul(dFnduR,duRduL)) * (third*bn_mag(i))

    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------
    ! End of two other nodes
    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------

    end do nodes_of_triangle

  end do boundary_element_loop

  endif

!--------------------------------------------------------------------------------
! Quadrialteral boundary faces:
!
! Close the residual by the 2nd-order boundary flux quadrature.
!
! Loop over quad boundary elements and distribute a contribution to each node. 
!               
!      o------------o
!       \            \
!        \            \ 
!         \            \
!          \            \ 
!           o------------o
!           j
!
!
!    Flux to be added to the residual at j is just
!
!      Flux(uj,ub)
!
!    where ub is a BC state.
!
!
! - This is simple for hexa grids. We only need to compute the flux at the node j and
!   add it to the residual at j. This is second-order accurate for regular hexa grids
!   with general rectangular cells. If a hexahedral cell is irregular with non-planar
!   faces, then second-order will be lost. See the reference below for details.
!
! - Second-order quadrature formulas for 3D tet/prs/pyr/hex grids can be found in:
!
! H. Nishikawa, "Beyond Interface Gradient: A General Principle for
! Constructing Diffusion Schemes", AIAA Paper 2010-5093, 40th AIAA Fluid
! Dynamics Conference and Exhibit, Chicago, 2010.
! http://ossanworld.com/hiroakinishikawa/My_papers/nishikawa_AIAA-2010-5093.pdf
! (NOTE: See Appendix B for boundary flux quadratures for 2D/3D elements.)
!
!--------------------------------------------------------------------------------

  if (nquad > 0) then

  boundary_element_loop_quad : do i = 1, nquad

    ib = quad(i,5)

    nodes_of_quad : do j = 1, 4

    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------
    ! Second-order accurate boundary flux quadratue for hexa (1,0,0,0)
    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------

     ! Node j: Phi_jb(bn): BC is implemented here.

      nj = quad( i, j )

      call set_bc_state( u(nj,1),u(nj,2),u(nj,3),u(nj,4),ib,bn(i,:),x(nj),y(nj),z(nj), ub,qnb) !Get ub,qnb
      call get_right_state(  bc_type(ib),u(nj,:),ub,qnb, bn(i,:),  duRduL,uR )                 !Get duR/duL
      call interface_jac_b( u(nj,:), uR, bn(i,:), sqrt(bn_mag(i))/ar(nj), dFnduL, dFnduR )

     !Contribution to node j (diagonal block):
      jac_diag(nj,:,:) = jac_diag(nj,:,:) + (dFnduL + matmul(dFnduR,duRduL)) * (third*bn_mag(i))

     ! That's it for hexa grids. Simple.

    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------
    ! End of Second-order accurate boundary flux quadratue for hexa (1,0,0,0)
    !-----------------------------------------------------------------------------------
    !-----------------------------------------------------------------------------------

    end do nodes_of_quad

  end do boundary_element_loop_quad

  endif
